/* llm - litle lisp machine:
 *
 *  Copyright (c) 2019  Christian Barthel <bch@online.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * Compile: flex tokenize.l
 *          cc lex.yy.c -lfl
 */
%{
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
struct token;
struct token {
  int           type;
  union {
    char       *str;
    int         num;
  } v;
  struct token *next;
};

struct ast; 
struct ast {
  int             type;
  union {
    struct token *token;
    struct ast   *list;
  } v;
  struct ast     *next;
};

struct env;
struct env {
  char       *name;
  struct ast *a;
  struct env *parent;
  struct env *next;
};

enum asttype {
  AST_TOK  = 400,
  AST_LIST = 401,
  AST_QUOTE = 402
};
enum yytokentype {
  NUM    = 258,
  LPAR   = 259,
  RPAR   = 260,
  SYM    = 261,
  STR    = 262,
  EOL    = 263,
  BOOL   = 264
};

struct token* make_token(enum yytokentype, int, char*);
struct ast*   eval(struct ast *a, struct env *e);
int   yylval;
char *yystr;
%}

%%
"("                  { return LPAR; }
")"                  { return RPAR; }
t                    { yylval = 1; return BOOL;}
nil                  { yylval = 0; return BOOL; }
[0-9]+               { yylval = atoi(yytext); return NUM; }
[ \t\n]              { /* ignore white space */ }
\".*\"               { return STR; }
[a-zA-Z][a-zA-Z0-9]* { return SYM; }
.                    { err(1, "invalid symbol: %s\n", yytext); }
%%

/*
 * Goal: evaluate simple LISP forms in an
 *       read-eval-print stylized loop.  supports:
 *
 *  Data types:
 *   #       Functions
 *   t,nil   boolean        (nand <bool> <bool>)      -> boolean
 *   "str"   Strings        (streq <string> <string>) -> boolean
 *   abc     Symboles       (symeq <sym> <sym>)       -> boolean
 *   1234    Number/Integer (inv <num>)               -> -num
 *                          (add <num> <num>)         -> num
 *                          (lt <num> <num>)          -> boolean
 *                          (num2bool <num>)          -> boolean
 *
 *   Special Forms:
 *     (def <var> <symbol,number,#function)
 *           Bind symbol <var> in the current environment to
 *           the evaluated form.
 *     (lm (<symbol> ..) (<sym..> ))
 *     (if <bool> <body> <else>)
 *     (quote a) -> a
 */

void
debug_token(struct token *x, int indent) {
  for (int i = 0; i < indent; i++) putc(' ', stderr);
  fprintf(stderr, "%d: (%p)", x->type, x);
  if (x->type == NUM) fprintf(stderr, "%d @ %p", x->v.num, x);
  if (x->type == STR || x->type == SYM)
    fprintf(stderr, "%s", x->v.str);
  putc('\n', stderr);
}

void
debug_ast(struct ast *a, int indent)
{
  while (a) {
    for (int i = 0; i < indent; i++) putc(' ', stderr);
    fprintf(stderr, "AST: %s,a=%p,list=%p,token=%p,next=%p", (a->type == AST_TOK) ? "AST_TOK" :
            "AST_LIST", a, a->v.list, a->v.token, a->next);
    putc('\n', stderr);
    if (a->type == AST_LIST) {
      fprintf(stderr, "enter %p\n", a->v.list);
      fprintf(stderr, "==\n");
      debug_ast(a->v.list, indent+4);
      fprintf(stderr, "==\n");
    }
    else if(a->type == AST_TOK)
      debug_token(a->v.token, indent);
    a = a->next;
  }
}

struct token*
make_token(enum yytokentype type, int num, char* str)
{
  struct token *t = 
      (struct token*) calloc(1, sizeof(struct token));
  if (t == NULL)
    err(1, "malloc failed");
  t->type = type;
  if (type == NUM || type == BOOL)
    t->v.num = num;
  else if (type == SYM || type == STR)
    t->v.str = strdup(str);
  return t;
}

struct ast*
make_ast(enum asttype type, struct token *t)
{
  struct ast *a = 
    (struct ast*) calloc(1, sizeof(struct ast));
  if (a == NULL)
    err(1, "malloc failed");
  a->type = type;
  if (type == AST_TOK)
    a->v.token = t;
  /* else if (type == AST_LIST) */
  /*   a->v.list = n;  */
  return a;
}

struct env*
make_env(char *name, struct ast *a)
{
   struct env *e = 
    (struct env*) calloc(1, sizeof(struct env));
   if (e == NULL)
     err(1, "malloc failed");
   e->name = name;
   e->a = a;
   return e;
}

int token_is_num(struct token *t) {return t->type == NUM;}
int token_is_sym(struct token *t) {return t->type == SYM;}
int token_is_str(struct token *t) {return t->type == STR;}
int token_is_bool(struct token *t) {return t->type == BOOL;}
int token_is_internal(struct token *t) {
  if (strcmp(t->v.str, "def") == 0) return 1;
  else if (strcmp(t->v.str, "lm") == 0) return 1;
  else if (strcmp(t->v.str, "if") == 0) return 1;
  else if (strcmp(t->v.str, "inv") == 0) return 1;
  else if (strcmp(t->v.str, "add") == 0) return 1;
  else if (strcmp(t->v.str, "lt") == 0) return 1;
  else if (strcmp(t->v.str, "nand") == 0) return 1;
  else if (strcmp(t->v.str, "q") == 0) return 1;
  else if (strcmp(t->v.str, "num2bool") == 0) return 1;
  else return 0;
}


/* upon parsing a list of tokens like:
 *    "(" -> "def" -> "(" -> "b" -> ")" -> "c" -> ")"
 * care must be taken when entering a new sub-list (b),
 * because the processing in the upper level must continue 
 * where the sub-list ended, i.e. "c" must be the next 
 * token on this level.
 */
struct token *next;
struct ast*
parse(struct token *t)
{
  if (t == NULL)
    err(1, "Unexpected token: <NULL>");

 if (t->type == LPAR) {
    struct ast *head = make_ast(AST_LIST, NULL);
    struct ast *a, *p;
    t = t->next;
    while (t->type != RPAR) {
      assert (t != NULL);
      a = parse(t);
      if (a == NULL)
        err(1, "syntax error");
      if (head->v.list == NULL)
        p = head->v.list = a;
      else
        p->next = a;
      p = a;
      t = next; /* skip all tokens that are processed */
      assert(t != NULL);
    }
    assert(t->type == RPAR);
    t = t->next;
    next = t;
    return head;
  } else if (t->type == NUM) {
    next = t->next;
    return make_ast(AST_TOK, t);
  } else if (t->type == BOOL) {
    next = t->next;
    return make_ast(AST_TOK, t);
  } else if (t->type == SYM) {
    next = t->next;
    return make_ast(AST_TOK, t);
  } else if (t->type == STR) {
    next = t->next;
    return make_ast(AST_TOK, t);
  }
  return NULL;
}

void
append(struct env *e, struct env *a)
{
  if (e->next == NULL) {
    e->next = a;
    return;
  } else 
    append(e->next, a);
}

struct ast *
lookup(struct env *e, char *name)
{
  if (e == NULL)
    return NULL;
  for (struct env *t = e; t != NULL; t = t->next)
    if (strcmp(t->name, name) == 0)
      return t->a;
  return lookup(e->parent, name);
}

/* (def a <ausdruck>) */
struct ast *
eval_def(struct ast *a, struct env *e)
{
  a = a->next;                        /* skrip `def` */
  assert((a->next->type == AST_TOK && 
          a->next->v.token->type== STR)  ||
         (a->next->type == AST_TOK &&
          a->next->v.token->type == BOOL) ||
         (a->next->type == AST_TOK && 
          a->next->v.token->type == NUM) ||
         (a->next->type == AST_LIST)     ||      
         (a->next->type == AST_TOK && 
          a->next->v.token->type== SYM));
  
  append(e, make_env(a->v.token->v.str, eval(a->next, e)));

  assert(a->next->next == NULL);
  return NULL;
}

/* (if <cond:bool> <expression> <expression>) */
struct ast *
eval_if(struct ast *a, struct env *e)
{
  a = a->next;

  assert(a != NULL && a->next != NULL);

  struct ast *condition = eval(a, e);
  
  assert(condition != NULL &&
         condition->type == AST_TOK &&
         condition->v.token->type == BOOL);
  
  assert(a->next != NULL &&
         a->next->next != NULL);

  if (condition->v.token->v.num == 1)
    return eval(a->next, e);
  else
    return eval(a->next->next, e);
}

/* (nand <ausdruck::bool> <audsruck::bool>) -> BOOL */
struct ast *
eval_nand(struct ast *a, struct env *e)
{
  assert (a != NULL && 
          a->next != NULL &&
          a->next->next != NULL &&
          a->next->next->next == NULL);

  struct ast *op1 = eval(a->next, e);
  struct ast *op2 = eval(a->next->next, e);
        
  assert (op1 != NULL &&
          op1->type == AST_TOK &&
          op1->v.token->type == BOOL &&
          op2 != NULL &&
          op2->type == AST_TOK &&
          op2->v.token->type == BOOL );
  return make_ast(AST_TOK, 
                  make_token(BOOL, 
                             !(
                             op1->v.token->v.num &
                             op2->v.token->v.num) ,
                             NULL));
}

/* (add <ausdruck::NUM> <audsruck::NUM>) -> NUM */
struct ast *
eval_add(struct ast *a, struct env *e)
{
  assert (a != NULL && 
          a->next != NULL &&
          a->next->next != NULL &&
          a->next->next->next == NULL);

  struct ast *op1 = eval(a->next, e);
  struct ast *op2 = eval(a->next->next, e);
        
  assert (op1 != NULL &&
          op1->type == AST_TOK &&
          op1->v.token->type == NUM &&
          op2 != NULL &&
          op2->type == AST_TOK &&
          op2->v.token->type == NUM);
  return make_ast(AST_TOK, 
                  make_token(NUM, 
                             op1->v.token->v.num + 
                             op2->v.token->v.num,
                             NULL));
}

/* (lt <ausdruck::NUM> <audsruck::NUM>) -> BOOL*/
struct ast *
eval_lt(struct ast *a, struct env *e)
{
  assert (a != NULL && 
          a->next != NULL &&
          a->next->next != NULL &&
          a->next->next->next == NULL);

  struct ast *op1 = eval(a->next, e);
  struct ast *op2 = eval(a->next->next, e);
        
  assert (op1 != NULL &&
          op1->type == AST_TOK &&
          op1->v.token->type == NUM &&
          op2 != NULL &&
          op2->type == AST_TOK &&
          op2->v.token->type == NUM);
  return make_ast(AST_TOK, 
                  make_token(BOOL, 
                             op1->v.token->v.num <
                             op2->v.token->v.num,
                             NULL));
}

/* (inv <ausdruck::NUM>) -> NUM */
struct ast *
eval_inv(struct ast *a, struct env *e)
{
  assert(a != NULL && 
         a->next != NULL && 
         a->next->next == NULL);
  struct ast *op = eval(a->next, e);
  assert(op != NULL &&
         op->type == AST_TOK &&
         op->v.token->type == NUM &&
         op->next == NULL);
  return make_ast(AST_TOK, 
                  make_token(NUM, 
                             0 - 
                             op->v.token->v.num,
                             NULL));
}

/* (num2bool <ausdruck::NUM>) -> bool */
struct ast *
eval_num2bool(struct ast *a, struct env *e)
{
  assert(a != NULL && 
         a->next != NULL && 
         a->next->next == NULL);
  struct ast *op = eval(a->next, e);
  assert(op != NULL &&
         op->type == AST_TOK &&
         op->v.token->type == NUM &&
         op->next == NULL);
  return make_ast(AST_TOK, 
                  make_token(BOOL, 
                             op->v.token->v.num != 0,
                             NULL));
}

struct ast *
eval_internal_sym(struct ast *a, struct env *e)
{
  if (strcmp(a->v.token->v.str, "def") == 0)
    return eval_def(a, e);
  else if (strcmp(a->v.token->v.str, "lm") == 0) {
  } else if (strcmp(a->v.token->v.str, "if") == 0) {
    return eval_if(a, e);
  } else if (strcmp(a->v.token->v.str, "inv") == 0) {
    return eval_inv(a, e);
  } else if (strcmp(a->v.token->v.str, "add") == 0) {
    return eval_add(a, e);
  } else if (strcmp(a->v.token->v.str, "lt") == 0) {
    return eval_lt(a, e);
  } else if (strcmp(a->v.token->v.str, "nand") == 0) {
    return eval_nand(a, e);
  } else if (strcmp(a->v.token->v.str, "q") == 0) {
    //return make_ast(AST_QUOTE, a->next);
    return a->next;
  } else if (strcmp(a->v.token->v.str, "num2bool") == 0) {
    return eval_num2bool(a, e);
  }
  return NULL;
}

struct ast *
eval_defined_sym(struct ast *a, struct env *e)
{
  struct ast *var = lookup(e, a->v.token->v.str);
  assert(var != NULL);            /* variable found? */
  return var;
}

struct ast *
eval(struct ast *a, struct env *e)
{
  if (a == NULL) return NULL;
  
  switch (a->type) {
  case AST_TOK:
    if (token_is_num(a->v.token))
      return make_ast(AST_TOK, a->v.token);
    else if (token_is_sym(a->v.token))
      return eval_defined_sym(a, e);
    else if (token_is_str(a->v.token))
      return make_ast(AST_TOK, a->v.token);
    else if (token_is_bool(a->v.token))
      return make_ast(AST_TOK, a->v.token);
  case AST_LIST:
    /* XXX either it's an internal function, a custom defined function
    *      or an lambda expression ((lm (x) (* x x)) 2) */
    assert ((a->v.list->type == AST_TOK &&
             a->v.list->v.token->type == SYM) ||
            (a->v.list->type == AST_LIST));
    if (token_is_internal(a->v.list->v.token))
      return eval_internal_sym(a->v.list, e);
    /* else if ((r = eval_defined_function(a->v.list, e)) != NULL) */
    /*   return r; */
    assert(0);
  default:
    fprintf(stderr, "%d\n", a->type);
    assert(0);
  }
  return NULL;
}

void
pr(struct ast *a)
{
  if (a == NULL) {
    fprintf(stderr, "null %p\n", a);
    return;
  }
  switch (a->type) {
  case AST_TOK:
    if (a->v.token->type == NUM)
      printf("%d\n", a->v.token->v.num);
    else if (a->v.token->type == STR)
      printf("%s\n", a->v.token->v.str);
    else if (a->v.token->type == SYM)
      printf("%s\n", a->v.token->v.str);
    else if (a->v.token->type == BOOL) {
      if (a->v.token->v.num)
        printf("t\n");
      else
        printf("nil\n");
    }
    break;
  case AST_LIST:
    printf("#fn\n");
  default:
    err(1, "don't know how to print");
  }
}

struct token*
read_form()
{
  int tok, open = 0;
  struct token *t, *u = NULL, *start = NULL;

  while((tok = yylex())) {
    if (tok == LPAR) open++;
    else if (tok == RPAR) open--;

    t = make_token(tok, yylval, yytext);
    if (start == NULL) start = t;
    if (u == NULL)     u = t;
    else               u->next = t;
    u = t;
    
    if (!open) {
      return start;
    }
  }
  exit(0);
}

int main(void)
{
  struct env default_env = { 0 }, *env;
  env = &default_env;
  env->name = "__dummy";
 
  while (1)
    pr(eval(parse(read_form()), env));
  return 0;
}


/*
 * Local Variables:
 * mode: c;
 * eval: (message "main()")
 * fill-column: 80
 * comment-column: 40
 * indent-tabs-mode: nil
 * tab-width: 2
 * c-basic-offset: 2
 * End:
 */
